# -*- perl -*-

# Copyright (c) 2007 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Created: 2007-Dec-22 12:27 (EST)
# Function: service for math calculations
#
# $Id: Argus::Compute.pm,v 1.3 2008/06/25 16:03:16 jaw Exp $

# But the age of chivalry is gone; that of sophisters, economists,
# and calculators has succeeded.
#   -- Edmund Burke, Reflections on the Revolution in France

package Argus::Compute;
@ISA = qw(Service);
use Argus::Encode;

use strict qw(refs vars);
use vars qw($doc @ISA);


$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(Service MonEl BaseIO)],
    methods => {},
    versn => '3.6',
    html  => 'compute',
    fields => {
      compute::expr => {
	  descr => 'mathmatical expression on other services',
	  exmpl => "( %{Top:Foo} + %{Top:Bar} ) / %{Top:Baz}",
	  attrs => ['config', 'inherit'],
      },
      compute::downtime => {},
      compute::srvcs => {},
    },
};

sub probe {
    my $name = shift;

    return [7, \&config] if $name =~ /^Compute/i;
}

sub config {
    my $me = shift;
    my $cf = shift;
    
    $me->init_from_config( $cf, $doc, 'compute' );

    my $expr = $me->{compute}{expr};
    $me->{uname} = "COMPUTE_$expr";

    # What mad expressions did my tongue refuse
    #   -- Virgil, Aenid
    return $cf->error( 'expr not specified' )
	unless $expr;
    
    bless $me if( ref($me) eq 'Service' );
    $me;
}

sub resolve_depends {
    my $me = shift;
    my $cf = shift;

    my $expr = $me->{compute}{expr};
    
    # convert %{Top:Foo} => %{Top:Foo{srvc::result}}
    $expr =~ s/%\{([^{}]+)\}/%\{$1\{srvc::result\}\}/g;

    $me->{compute}{expr} = $expr;

    # get list of services
    my @srvc = $expr =~ /%\{([^{}]+)/g;

    unless( $me->{compute}{srvcs} ){
	for my $s (@srvc){
	    my $x = $MonEl::byname{$s};
	    
	    unless( $x ){
		$me->warning("Cannot resolve compute dependency: $s");
		next;
	    }
	    
	    # She that would alter services with thee
	    #   -- Shakespeare, Twelfth Night
	    # add self to srvc, + v/v
	    push @{$x->{srvc}{alsorun}}, $me;
	    push @{$me->{compute}{srvcs}}, $x;
	}
    }
   
    $me->SUPER::resolve_depends($cf);
}

# Let ignorance talk as it will, learning has its value.
#   -- Jean de La Fontaine, The Use of Knowledge
sub value {
    my $me = shift;
    my $s  = shift;
    my $p  = shift;

    my $x = $MonEl::byname{$s};
    unless($x){
	# Does not compute!
	#   -- Robot, Lost In Space
	$me->debug("does not compute. object not found: $s");
	return "#<invalid object>";
    }
    my $r = eval {
	$x->getparam($p) || 0;
    };
    if(my $e = $@){
	$me->debug("does not compute. param not found: $s\{$p\}");
	return "#<invalid param>";
    }

    $me->debug("computing $s\{$p\} => $r");

    $r;
}

sub start {
    my $me = shift;
    
    $me->SUPER::start();
    $me->debug( 'start compute' );

    # are all srvcs ready?

    for my $s ( @{$me->{compute}{srvcs}} ){
	unless( $s->{srvc}{result_valid} ){
	    $me->debug("service not ready: " . $s->unique() );
	    return $me->done();
	}
    }
    
    my $expr = $me->{compute}{expr};

    # fill in values
    $expr =~ s/%\{([^{}]+)\{([^{}]+)\}\}/value($me,$1,$2)/ge;
    $me->debug("computing: $expr");
    my $r = eval $expr;
    
    return $me->isdown("computation failed: $@")  if $@;

    # Is wander'd forth, in care to seek me out
    # By computation and mine host's report.
    #   -- Shakespeare, Comedy of Errors
    
    $me->generic_test($r, 'COMPUTE');

}

sub isup {
    my $me = shift;

    $me->{compute}{downtime} = undef;
    $me->SUPER::isup(@_);
}

sub isdown {
    my $me = shift;

    # emulate retries in a reasonable way
    if( $me->{srvc}{retries} ){
	$me->{compute}{downtime} ||= $^T;
	my $t = 1 + ($^T - $me->{compute}{downtime}) / $me->{srvc}{frequency};
	$me->{srvc}{tries} = $t;
    }
    
    $me->SUPER::isdown(@_);
}

sub check_now {
    my $me = shift;

    # check all component srvcs
    for my $s ( @{$me->{compute}{srvcs}} ){
	$s->check_now();
    }
}

sub recycle {
    my $me = shift;

    delete $me->{compute}{srvcs};
    $me->SUPER::recycle(@_);
}

sub reschedule {
    # not scheduled
    # will run when underlying services run
}

sub about_more {
    my $me  = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->SUPER::about_more($ctl);
    $me->more_about_whom($ctl, 'compute');
}


################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;

